# coding=utf8
# Copyright 2020 Cognicept Systems
# Author: Swarooph Seshadri (swarooph@cognicept.systems)
# --> Pusher class handles pushing stuff to the Cognicept cloud

import os
import sys
import time
import glob
import requests
import datetime
import threading
import ntpath
import boto3
from boto3.s3.transfer import TransferConfig
from cogniceptshell.common import bcolors


class ProgressPercentage(object):
    """
    A class to track upload progress for chunks uploaded to S3
    ...

    Parameters
    ----------
    _filename (str): Name of the file that is being uploaded
    _size (float): Size of the file being uploaded
    _seen_so_far (float): Size of the file already uploaded
    _lock (threading.lock): Thread lock object for upload context

    Methods
    -------
    __init__(filename):
        Constructor for ProgressPercentage class that initializes all the parameters.
    __call__(bytes_amount):
        Main method to print progress percentage of file being uploaded.
    """

    def __init__(self, filename):
        """
        Constructor for ProgressPercentage class that initializes all the parameters.

                Parameters:
                        filename: Name of the file that is being uploaded
        """
        self._filename = filename
        self._size = float(os.path.getsize(filename))
        self._seen_so_far = 0
        self._lock = threading.Lock()

    def __call__(self, bytes_amount):
        """
        Main method to print progress percentage of file being uploaded.

                Parameters:
                        bytes_amount: Size of the file already uploaded
        """
        # To simplify, assume this is hooked up to a single filename
        with self._lock:
            self._seen_so_far += bytes_amount
            percentage = (self._seen_so_far / self._size) * 100
            sys.stdout.write(
                "\r%s  %s / %s  (%.2f%%)" % (
                    self._filename, self._seen_so_far, self._size,
                    percentage))
            sys.stdout.flush()


class Pusher:
    """
    A class to manage Cognicept's file pushes to the Cognicept cloud
    ...

    Parameters
    ----------
    bucket_name (str): Name of the S3 bucket to upload to

    Methods
    -------
    __init__():
        Constructor for Pusher class that initializes all the parameters.
    push(args):
        Main entry point for the push operation.
        Checks arguments and if valid, calls other methods to construct client and push the file.
    construct_client(config_obj):
        Creates an S3 client for temporary credentials fetched by key rotate operation.
        Uses an instance of Configuration class.
    push_bag(args):
        Pushes bag to S3 and posts bag metadata to Cognicept cloud.
    get_latest_bag(args):
        Utility method to get the latest '.bag' file.
    check_bag_exists(bag_file_path):
        Utility method to check if specified bag file exists.
    upload_s3(bag_file_path, config_obj):
        Uploads specified bag file to configured S3 bucket.
        Has built in retry mechanism  in case upload fails.
    post_bag_metadata(bag_file_path):
        Posts bag metadata to the Cognicept Cloud using Cognicept API.
    """
    def __init__(self):
        """
        Constructor for Pusher class that initializes all the parameters.

                Parameters:
                        bucket_name: Name of the S3 bucket to upload to
        """
        self.bucket_name = 'cognicept-bagfiles'

    def push(self, args):
        """
        Main entry point for the push operation.
        Checks arguments and if valid, calls other methods to construct client and push the file.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
        """
        if args.bag:
            # Construct client
            self.construct_client(args.config)
            # Push bag
            self.push_bag(args)
        else:
            # start or stop must be provided
            print(
                bcolors.FAIL + "Required command is missing." + bcolors.ENDC)

    def construct_client(self, config_obj):
        """
        Creates an S3 client for temporary credentials fetched by key rotate operation.

                Parameters:
                        config_obj: An instance of Configuration class.
        """
        rotate_status = config_obj.cognicept_key_rotate(None)

        if rotate_status:
            local_cfg = config_obj.config
            self.s3_client = boto3.client(
                's3', region_name='ap-southeast-1',
                aws_access_key_id=local_cfg['AWS_ACCESS_KEY_ID'],
                aws_secret_access_key=local_cfg['AWS_SECRET_ACCESS_KEY'],
                aws_session_token=local_cfg['AWS_SESSION_TOKEN'])
        else:
            raise(SystemExit)

    def push_bag(self, args):
        """
        Pushes bag to S3 and posts bag metadata to Cognicept cloud.
        If `args.bag` is set to `latest`, automatically finds the latest file to upload.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
        """
        upload_status = False
        if args.bag == 'latest':
            bag_file_path = self.get_latest_bag(args)
            print(
                bcolors.OKBLUE +
                "Uploading latest bag file to S3: " +
                bcolors.ENDC + bag_file_path)
            upload_status = self.upload_s3(bag_file_path, args.config)
        elif args.bag.endswith('.bag'):
            bag_file_path = os.path.expanduser(args.path+"bags/"+args.bag)
            if(self.check_bag_exists(bag_file_path)):
                print(
                    bcolors.OKBLUE +
                    "Uploading bag file to S3: " +
                    bcolors.ENDC + bag_file_path)
                upload_status = self.upload_s3(bag_file_path, args.config)
            else:
                print(
                    bcolors.FAIL +
                    "Specified bag file not found in expected location: " +
                    bcolors.ENDC + bag_file_path)
        else:
            print(
                bcolors.FAIL + "Only .bag files can be uploaded." + bcolors.ENDC)

        if upload_status:
            print(bcolors.OKGREEN + "\nBag file uploaded." + bcolors.ENDC)
            self.post_bag_metadata(args, bag_file_path)
        else:
            pass

    def get_latest_bag(self, args):
        """
        Utility method to get the latest '.bag' file. 
        If no file is found, exits program.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        latest_file (string): Full path of the latest bag file
        """
        bag_path = os.path.expanduser(args.path+"bags")
        bag_list = glob.glob(bag_path+"/*.bag")
        if bag_list:
            latest_file = max(bag_list, key=os.path.getctime)
            return latest_file
        else:
            print(
                bcolors.FAIL + "No bags found in the expected path." + bcolors.ENDC)
            raise(SystemExit)

    def check_bag_exists(self, bag_file_path):
        """
        Utility method to check if specified bag file exists.

                Parameters:
                        bag_file_path: Full path of the bag file to check
                Returns:
                        exist_status (boolean): True if specified bag file exists
        """
        if (os.path.exists(bag_file_path)):
            return True
        else:
            return False

    def upload_s3(self, bag_file_path, config_obj):
        """
        Uploads specified bag file to configured S3 bucket.
        Has built in retry mechanism  in case upload fails.

                Parameters:
                        bag_file_path: Full path of the bag file to upload
                        config_obj: An instance of Configuration class
                Returns:
                        upload_status (boolean): True if bag file is uploaded successfully
        """
        # Multipart threaded approach for larger files
        config = TransferConfig(multipart_threshold=1024*25, max_concurrency=10,
                                multipart_chunksize=1024*25, use_threads=True)
        num_retries = 3

        for trial in range(num_retries):
            try:
                # Upload file
                file_name = ntpath.basename(bag_file_path)
                self.s3_client.upload_file(bag_file_path, self.bucket_name, file_name,
                                           Config=config, Callback=ProgressPercentage(bag_file_path))
                # Return true for successful upload
                return True
            except boto3.exceptions.S3UploadFailedError:
                # On failure, retry
                print('Attempt #' + str(trial+1) + bcolors.FAIL +
                      " FAILED" + bcolors.ENDC + "\033[K")
                # Wait for 1 second before retrying
                time.sleep(1.0)
        # If the loop is completed, upload failed, so return false
        return False

    def post_bag_metadata(self, args, bag_file_path):
        """
        Posts bag metadata to the Cognicept Cloud using Cognicept API.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                        bag_file_path: Full path of the bag file to upload
                Returns:
                        resp_status (boolean): True if metadata posted successfully
        """
        payload = {
            "robot_id": args.config.get_field("ROBOT_CODE"),
            "bagfile_name": ntpath.basename(bag_file_path),
            "bagfile_url": ntpath.basename(bag_file_path),
            "bagfile_size": str(os.path.getsize(bag_file_path)),
            "uploaded_at": datetime.datetime.now().isoformat()
        }

        # retrieve cognicept access key
        try:
            token = args.config.get_field("COGNICEPT_ACCESS_KEY")
        except KeyError:
            print(bcolors.FAIL +
                  'COGNICEPT_ACCESS_KEY variable undefined in configuration. Could not update Cognicept cloud with bag metadata.' +
                  bcolors.ENDC)
            raise(SystemExit)

        headers = {"Authorization": "Basic " + token}

        try:
            resp = requests.post(args.config.get_cognicept_api_uri(
            ) + "bagfile", headers=headers, json=payload, timeout=5)

            if resp.status_code != 200:
                print(
                    'Cognicept REST API error: ' +
                    args.config.get_cognicept_api_uri() +
                    ' responded with ' + str(resp.status_code) +
                    '\n' + resp.json()['Message'])
                return False
            else:
                print(
                    bcolors.OKGREEN +
                    "Metadata posted to Cognicept Cloud." +
                    bcolors.ENDC)
                return True
        except requests.exceptions.Timeout:
            print("Cognicept REST API error: time out.")
            return False
        except requests.exceptions.TooManyRedirects:
            print("Cognicept REST API error: Wrong endpoint.")
            return False
        except Exception as ex:
            print("Cognicept REST API error, " + str(ex))
            return False



# Copyright 2020 Cognicept Systems
# Author: Jakub Tomasek (jakub@cognicept.systems)
# --> AgentLifeCycle handles life cycle of Cognicept Agents

import time
import docker
import boto3
import base64
import json
import os
import sys
import subprocess
import dateutil
from datetime import datetime
import re
import glob
from multiprocessing import Process, Queue
import requests
import botocore
import shutil
from cogniceptshell.common import bcolors
from cogniceptshell.common import generate_progress_bar
from cogniceptshell.common import permission_safe_docker_call
from subprocess import DEVNULL
from docker.types import LogConfig


class AgentLifeCycle:
    """
    A class to manage agent and Cognicept's docker container lifecycle
    ...

    Parameters
    ----------
    None

    Methods
    -------
    configure_containers(cfg):
        Loads agent configuration from `COG_AGENT_CONTAINERS` and `COG_AGENT_IMAGES`
    get_status(args):
        Prints status of docker containers listed in `COG_AGENT_CONTAINERS`.
    get_last_event(args):
        Prints last log in `~/.cognicept/agent/logs/`.
    restart(args):
        Stops and starts the containers listed in `COG_AGENT_CONTAINERS`.
    start(args):
        Starts the containers listed in `COG_AGENT_CONTAINERS`. If `args` has parameter `list`, starts only containers in the list.
    stop(args):
        Stops the containers listed in `COG_AGENT_CONTAINERS`. If `args` has parameter `list`, stops only containers in the list.
    update(args):
        Pulls docker images listed in `COG_AGENT_IMAGES`.
    run_orbitty(args):
        Starts Orbitty.
    """

    # default configuration of containers and images
    _docker_container_names = ["cgs_diagnostics_agent", "remote_intervention_agent",
                               "cgs_diagnostics_ecs_api", "cgs_diagnostics_streamer_api", "cgs_bagger_server"]
    _docker_images = {}
    _docker_images["remote_intervention_agent"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/remote_intervention_agent:latest"
    _docker_images["kriya_watchdog"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/remote_intervention_agent:latest"
    _docker_images["cgs_diagnostics_agent"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/cognicept_diagnostics_agent:latest"
    _docker_images["cgs_diagnostics_ecs_api"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/cognicept_diagnostics_api:latest"
    _docker_images["cgs_diagnostics_streamer_api"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/cognicept_diagnostics_api:latest"
    _docker_images["orbitty"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/orbitty:latest"
    _docker_images["cgs_bagger_server"] = "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/cognicept_rosbagger:latest"

    def configure_containers(object, cfg):
        """
        Loads agent configuration from `COG_AGENT_CONTAINERS` and `COG_AGENT_IMAGES`

                Parameters:
                        cfg (Configuration): Cognicept configuration
                Returns:
                        None
        """

        if("COG_AGENT_CONTAINERS" in cfg.config and "COG_AGENT_IMAGES" in cfg.config):
            container_names = cfg.config["COG_AGENT_CONTAINERS"].split(";")
            image_names = cfg.config["COG_AGENT_IMAGES"].split(";")
            if(len(image_names) == len(container_names)):
                object._docker_container_names = container_names
                object._docker_images = {}
                i = 0
                for container_name in container_names:
                    object._docker_images[container_name] = image_names[i]
                    i = i + 1
            else:
                print(
                    "`COG_AGENT_CONTAINERS` and `COG_AGENT_IMAGES` do not coincide. Using default.")

            if("COG_ORBITTY_ENABLED" in cfg.config and "COG_ORBITTY_IMAGE" in cfg.config):
                if(bool(cfg.config["COG_ORBITTY_ENABLED"])):
                    object._docker_images["orbitty"] = cfg.config["COG_ORBITTY_IMAGE"]
        else:
            print(
                "Undefined `COG_AGENT_CONTAINERS` or `COG_AGENT_IMAGES`. Using default.")

    def _get_latest_log_loc(object, args):
        """
        Retrieve path to the last log in `~/.cognicept/agent/logs/` relative to `~/.cognicept/` or `path` specified by args.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        latest_log_loc (str): path to latest log relative to `~/.cognicept/`
        """
        # get latest log location
        latest_log_loc_file_path = os.path.expanduser(
            args.path+"agent/logs/latest_log_loc.txt")
        latest_log_loc = ""
        try:
            with open(latest_log_loc_file_path) as txt_file:
                latest_log_loc_temp = txt_file.readline()
                latest_log_loc_temp = latest_log_loc_temp[:-1]
                latest_log_loc = latest_log_loc_temp.replace(
                    "/$HOME/.cognicept/", "")
                latest_log_loc = latest_log_loc.replace(".cognicept/", "")
        except:
            cgs_agent_status = bcolors.FAIL + "UNKNOWN" + bcolors.ENDC

        return latest_log_loc

    def get_status(object, args):
        """
        Prints status of docker containers listed in `COG_AGENT_CONTAINERS`.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`

        """
        permission_docker_call_result = permission_safe_docker_call(docker.from_env)
        if permission_docker_call_result is None:
            return False
        client = permission_docker_call_result
        # check status of cgs_agent
        # get latest log location
        latest_log_loc = object._get_latest_log_loc(args)
        # read latest status and set for display
        file_path = os.path.expanduser(
            args.path+latest_log_loc+"/logDataStatus.json")
        try:
            with open(file_path) as json_file:
                data = json.load(json_file)
                period_since_update = datetime.utcnow(
                ) - dateutil.parser.parse(data["timestamp"])
                if(period_since_update.seconds < 30 and period_since_update.seconds >= 0):
                    cgs_agent_status = bcolors.OKBLUE + \
                        data["message"].upper() + bcolors.ENDC
                else:
                    cgs_agent_status = bcolors.WARNING + "STALE" + bcolors.ENDC

        except:
            cgs_agent_status = bcolors.FAIL + "Error" + bcolors.ENDC

        for container_name in object._docker_container_names:
            print(container_name, end=': ', flush=True)
            try:
                container = client.containers.get(container_name)
                if container.status != "running":
                    print(bcolors.WARNING + "OFFLINE" + bcolors.ENDC)
                else:
                    if(container_name == "cgs_diagnostics_agent"):
                        print(cgs_agent_status)
                    elif(container_name == "remote_intervention_agent"):
                        object._parse_remote_intervention_agent_logs(
                            container.logs(tail=50))
                    else:
                        print(bcolors.OKBLUE + "ONLINE" + bcolors.ENDC)
            except docker.errors.NotFound:
                print(bcolors.FAIL + "CONTAINER NOT FOUND" + bcolors.ENDC)

    def get_last_event(object, args):
        """
        Prints last log in `~/.cognicept/agent/logs/`.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`

        """
        # get latest log location
        latest_log_loc = object._get_latest_log_loc(args)

        # read and display latest log if any
        file_path = os.path.expanduser(args.path+latest_log_loc)
        try:
            print(bcolors.OKBLUE +
                  "Looking for the latest event log..." + bcolors.ENDC)
            # should read only latest logData#.json file and not logDataStatus.json
            list_of_log_files = [fn for fn in glob.glob(file_path + "/*.json")
                                 if not os.path.basename(fn).endswith("logDataStatus.json")]
            latest_log_file = max(list_of_log_files, key=os.path.getctime)

            with open(latest_log_file) as json_file:
                data = json.load(json_file)
                print(bcolors.OKGREEN+"Latest Event Log:" + bcolors.ENDC)
                print(json.dumps(data, indent=4, sort_keys=True))
        except:
            print(bcolors.WARNING + "No event logs present." + bcolors.ENDC)

    def _parse_remote_intervention_agent_logs(object, logs):
        """
        Parses logs to find status of remote intervention agent. Prints status.

                Parameters:
                        logs: container logs

        """
        logs_lines = logs.splitlines()
        # parse logs to get current status
        ri_agent_status = {}
        ri_agent_status["AGENT"] = ""
        ri_agent_status["WEBRTC"] = ""
        ri_agent_status["WEBSOCKET"] = ""

        # find latest status of the each module (agent, webrtc, websocket)
        for line in reversed(logs_lines):
            for key, value in ri_agent_status.items():
                if(value != ""):
                    continue
                matches = re.search(
                    '^.*{}:: STATUS:: (?P<status>.*).*$'.format(key), str(line))
                if(matches is not None):
                    ri_agent_status[key] = matches.groups(0)[0]
            if(ri_agent_status["AGENT"] != "" and ri_agent_status["WEBRTC"] != "" and ri_agent_status["WEBSOCKET"] != ""):
                continue

        output_text = bcolors.OKBLUE + "ONLINE" + bcolors.ENDC

        for key, value in ri_agent_status.items():
            if(value == ""):
                # if not found, it was not yet initialized
                output_text = bcolors.WARNING + "NOT INITIALIZED" + bcolors.ENDC
                break
            if(value != "OK"):
                output_text = bcolors.WARNING + key + value + bcolors.ENDC
                break
        print(output_text)

    def _detached_restart(object,args):
        """
        Runs _restart_protocol in detached mode with printing muted.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """

        if os.fork() != 0:
            return
        sys.stdout = open(os.devnull, 'w')
        result = object._restart_protocol(args)
        sys.stdout = sys.__stdout__
        return result


    def _restart_protocol(object,args):
        """
        Stops and starts the containers listed in `COG_AGENT_CONTAINERS`.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """
        object.stop(args)
        result = object.start(args)
        return result

    def restart(object, args):
        """
        Stops and starts the containers listed in `COG_AGENT_CONTAINERS`.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded in attached mode, always True in detached mode
        """
        if args.detach:
            result = True
            print("Running restart in detached mode")
            p = Process(target=object._detached_restart,args=(args,))
            p.start()
        else:
            result = object._restart_protocol(args)

        return result

    def start(object, args):
        """
        Starts the containers listed in `COG_AGENT_CONTAINERS`. If `args` has parameter `list`, starts only containers in the list.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """
        print("Called start")
        print("Starting agents")
        result = object.run_agents(args)
        return result

    def stop(object, args):
        """
        Stops the containers listed in `COG_AGENT_CONTAINERS`. If `args` has parameter `list`, stops only containers in the list.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """

        print("Stopping agents")
        result = object.remove_agents(args)
        return result

    def status(object, args):
        object.get_status(args)
        object.status_datadog(args)
        return True

    def remove_agents(object, args):
        """
        Stops the containers listed in `args.list`.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """

        # if list of agents is not specified, restart all
        if(not hasattr(args, 'list') or len(args.list) == 0):
            args.list = object._docker_container_names
        permission_docker_call_result = permission_safe_docker_call(docker.from_env)
        if permission_docker_call_result is None:
            return False

        client = permission_docker_call_result
        print("STOP: ")
        flag_success = True
        for container_name in args.list:
            print("   - " + container_name, end=': ', flush=True)
            try:
                container = client.containers.get(container_name)
                container.stop(timeout=10)
                container.remove()
                print(bcolors.OKBLUE + "DONE" + bcolors.ENDC)
            except docker.errors.NotFound:
                print(bcolors.WARNING + "NOT FOUND" + bcolors.ENDC)
                flag_success = False
            except docker.errors.APIError:
                print(bcolors.FAIL + "ERROR" + bcolors.ENDC)
                flag_success = False
        return flag_success

    def run_agents(object, args):
        """
        Starts the containers listed in `args.list`.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """
        # if list of agents is not specified, restart all
        if(not hasattr(args, 'list') or len(args.list) == 0):
            args.list = object._docker_container_names

        object._agent_run_options = {}
        object._agent_run_options["cgs_diagnostics_agent"] = {"command": "rosrun error_resolution_diagnoser error_resolution_diagnoser", "volumes": {
            args.config.config_path + "agent/logs/": {"bind": "/root/.cognicept/agent/logs", "mode": "rw"}}, "network_mode": "host"}
        object._agent_run_options["cgs_diagnostics_ecs_api"] = {
            "command": "/src/ecs_endpoint.py", "network_mode": "host"}
        object._agent_run_options["cgs_diagnostics_streamer_api"] = {"command": "/src/streamer_endpoint.py", "volumes": {
            args.config.config_path + "agent/logs/bunched_logs": {"bind": "/root/.cognicept/agent/logs/bunched_logs", "mode": "rw"}}, "network_mode": "host"}
        object._agent_run_options["remote_intervention_agent"] = {
            "command": "rosrun remote_intervention_agent cognicept_agent_node", "network_mode": "host"}

        if(args.config.is_ssh_enabled()):
            object._agent_run_options["remote_intervention_agent"]["volumes"] = {
                args.config.config_path + "ssh/id_rsa": {"bind": "/root/.ssh/id_rsa", "mode": "rw"}}
        object._agent_run_options["kriya_watchdog"] = {
            "command": "python3 /home/watchdog.py","volumes": {
            args.config.config_path + "kriya_logs/": {"bind": "/root/logs/", "mode": "rw"}, 
            "/var/run/docker.sock/": {"bind": "/var/run/docker.sock/", "mode": "rw"}}}
        object._agent_run_options["rosbridge"] = {
            "command": "roslaunch rosbridge_server rosbridge_websocket.launch", "network_mode": "host"}
        object._agent_run_options["test"] = {"command": "bash"}
        object._agent_run_options["cgs_bagger_server"] = {"command": "rosrun cognicept_rosbagger bagger_action_server.py", "volumes": {
            args.config.config_path + "bags/": {"bind": "/root/.cognicept/bags", "mode": "rw"}}, "network_mode": "host"}
        object._agent_run_options["other"] = {
            "command": "", "network_mode": "host"}
        permission_docker_call_result = permission_safe_docker_call(docker.from_env)
        if permission_docker_call_result is None:
            return False
        client = permission_docker_call_result
        print("RUN: ")
        success_flag = True
        for container_name in args.list:
            print("   - " + container_name, end=': ', flush=True)
            try:
                if(container_name not in object._docker_images):
                    if("COG_AGENT_CONTAINERS" in args.config.config):
                        containers = " (configured list: " + \
                            args.config.config["COG_AGENT_CONTAINERS"] + ")"
                    else:
                        containers = ""
                    print(bcolors.WARNING + "NOT FOUND" +
                          bcolors.ENDC + containers)
                    success_flag = False
                    continue

                if(container_name in object._agent_run_options.keys()):
                    options = object._agent_run_options[container_name]
                else:
                    options = object._agent_run_options["other"]
                options["name"] = container_name
                options["detach"] = True
                options["environment"] = args.config.config
                options["restart_policy"] = {"Name": "unless-stopped"}
                options["tty"] = True
                options["log_config"] = LogConfig(
                    type=LogConfig.types.JSON, config={'max-size': '5m'})
                if "command" in options:
                    command = options.pop("command")
                else:
                    command = ""
                container = client.containers.run(
                    object._docker_images[container_name], command, **options)
                print(bcolors.OKBLUE + "DONE" + bcolors.ENDC)
            except docker.errors.ContainerError:
                print(bcolors.WARNING + "ALREADY EXISTS" +
                      bcolors.ENDC + " (run `cognicept update`)")
            except docker.errors.ImageNotFound:
                print(bcolors.WARNING + "IMAGE NOT FOUND" +
                      bcolors.ENDC + " (run `cognicept update`)")
                success_flag = False
            except docker.errors.APIError:
                print(bcolors.FAIL + "DOCKER ERROR" + bcolors.ENDC)
                success_flag = False
        return success_flag

    def status_datadog(object, args):
        """
        Prints datadog status.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """
        sts = subprocess.call(
            ['sudo','sh','-c',"systemctl status datadog-agent", "-p"], shell=True, stdout=DEVNULL, stderr=DEVNULL)
        if sts == 0:
            print("Health monitor status:" +
                  bcolors.OKBLUE + " ACTIVE" + bcolors.ENDC)
        elif sts == 4:
            print("Health monitor status:" +
                  bcolors.FAIL + " NOT FOUND" + bcolors.ENDC)
        else:
            print("Health monitor status:" +
                  bcolors.WARNING + " INACTIVE" + bcolors.ENDC)

    def run_orbitty(object, args):
        """
        Starts Orbitty.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        None
        """
        os.system("xhost +local:root")
        permission_docker_call_result = permission_safe_docker_call(docker.from_env)
        if permission_docker_call_result is None:
            return False
        client = permission_docker_call_result
        try:
            options = {}
            options["name"] = "orbitty"
            options["detach"] = False
            options["privileged"] = True
            options["volumes"] = {}
            options["volumes"][args.config.config_path] = {
                "bind": "/config", "mode": "rw"}
            options["volumes"]["/tmp/.X11-unix"] = {
                "bind": "/tmp/.X11-unix", "mode": "rw"}
            environment = args.config.config
            environment["QT_X11_NO_MITSHM"] = 1
            environment["DISPLAY"] = ":0"
            options["environment"] = args.config.config
            options["remove"] = True
            options["tty"] = True
            command = "roslaunch orbitty orbitty.launch"
            client.containers.run(
                object._docker_images["orbitty"], command, **options)
        except docker.errors.ContainerError:
            print(bcolors.WARNING + "ALREADY RUNNING" + bcolors.ENDC)
        except docker.errors.ImageNotFound:
            print(bcolors.WARNING + "IMAGE NOT FOUND" +
                  bcolors.ENDC + " (run `cognicept update`)")
        except docker.errors.APIError:
            print(bcolors.FAIL + "DOCKER ERROR" + bcolors.ENDC)
        os.system("xhost -local:root")

    def _update_protocol(object, args):
        """
        Pulls docker images listed in `COG_AGENT_IMAGES`.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """
        # Attempt to login to ECR client
        login_success = object._ecr_login(args)

        # If login failed, return false
        if login_success is False:
            print("Your update credentials may have expired. Run `cognicept keyrotate` to refresh credentials and try `cognicept update` again.")
            return False

        print("Info: This may take a while depending on your connection.")
        images = set(object._docker_images.values())

        # load extra images to update
        if("COG_EXTRA_IMAGES" in args.config.config):
            image_names = args.config.config["COG_EXTRA_IMAGES"].split(";")
            if(len(image_names) > 0):
                images = images.union(set(image_names))

        N = len(images)
        i = 0
        success_flag = True
        for image_name in images:
            i = i + 1
            image_name_short = image_name.replace(
                "412284733352.dkr.ecr.ap-southeast-1.amazonaws.com/", "cognicept/")
            try:
                for status in object.docker_client.pull(image_name, stream=True, decode=True):
                    if("progress" not in status):
                        status["progress"] = ""
                    if("status" not in status):
                        status["status"] = "Error"
                    terminal_size = shutil.get_terminal_size().columns
                    progress = ""
                    if("progressDetail" in status and "total" in status["progressDetail"]):
                        progress = generate_progress_bar(
                            status["progressDetail"]["current"], status["progressDetail"]["total"], 1, 10)
                        status = "[" + str(i) + "/" + str(N) + "] " + image_name_short + \
                            " - " + status["status"] + " " + progress
                        if(terminal_size > 0):
                            print('{:{terminal_size}.{terminal_size}}'.format(
                                status, terminal_size=terminal_size), end="\r", flush=True)
                        else:
                            print('{:{trm_sz}.{trm_sz}}'.format(
                                status, trm_sz=80), end="\r", flush=True)
                print("[" + str(i) + "/" + str(N) + "] " + image_name_short +
                      " - " + bcolors.OKBLUE + "OK" + bcolors.ENDC + "\033[K")
            except docker.errors.ImageNotFound:
                print("[" + str(i) + "/" + str(N) + "] " + image_name_short +
                      " - " + bcolors.FAIL + "FAILED" + bcolors.ENDC + "\033[K")
                success_flag = False
            except:
                print("[" + str(i) + "/" + str(N) + "] " + image_name_short +
                      " - " + bcolors.FAIL + "FAILED" + bcolors.ENDC + "\033[K")
                success_flag = False

        print("Info: Run `cognicept restart` to redeploy updated agents.")
        return success_flag

    def _detached_update(object,args):
        """
        Runs the _update_protocol that pulls docker images listed in `COG_AGENT_IMAGES`.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """
        if os.fork() != 0:
            return

        sys.stdout = open(os.devnull, 'w')

        result = object._update_protocol(args)

        sys.stdout = sys.__stdout__

        return result

    def update(object,args):
        """
        Pulls docker images listed in `COG_AGENT_IMAGES`. Can be run in detached mode where printing is muted

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): Always True if run in detached mode, otherwise True if successful
        """
        if args.detach:
            result = True
            print("Running update in detached mode")
            p = Process(target=object._detached_update, args=(args,))
            p.start()
        else:
            result = object._update_protocol(args)

        return result
    def _construct_ecr_client(object, config_obj):
        """
        Member utility function to create ECR client `object.ecr_client` based on runtime.env credentials

                Parameters:
                        config_obj (Configuration): object holding Cognicept configuration
                Returns:
                        None      
        """
        # Get config
        local_cfg = config_obj.fetch_aws_keys()
        if local_cfg == False:
            return False
        if 'SessionToken' in local_cfg:
            object.ecr_client = boto3.client(
                'ecr', region_name='ap-southeast-1',
                aws_access_key_id=local_cfg['AccessKeyId'],
                aws_secret_access_key=local_cfg['SecretAccessKey'],
                aws_session_token=local_cfg['SessionToken'])
        else:
            object.ecr_client = boto3.client(
                'ecr', region_name='ap-southeast-1',
                aws_access_key_id=local_cfg['AccessKeyId'],
                aws_secret_access_key=local_cfg['SecretAccessKey'])
        return True

    def _ecr_login(object, args):
        """
        Member utility function that is called to login to ECR

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """
        # Construct client
        result = object._construct_ecr_client(args.config)
        if result == False:
            return False
        num_retries = 3
        for trial in range(num_retries):
            try:
                # Get token
                token = object.ecr_client.get_authorization_token()
                # Parse username, password
                username, password = base64.b64decode(
                    token['authorizationData'][0]['authorizationToken']).decode().split(':')
                # Get registry
                registry = token['authorizationData'][0]['proxyEndpoint']
                # Login
                docker_call_result = permission_safe_docker_call(docker.APIClient,
                                                                 base_url='unix://var/run/docker.sock')
                if docker_call_result is None:
                    return False

                object.docker_client = docker_call_result

                object.docker_client.login(username, password,
                                           registry=registry, reauth=True)
                # Return true for successful login
                return True
            except (docker.errors.APIError, botocore.exceptions.ClientError):
                # On failure, retry
                print('Attempt #' + str(trial+1) + bcolors.FAIL +
                      " FAILED" + bcolors.ENDC + "\033[K")
                # Wait for 1 second before retrying
                time.sleep(1.0)
        # If the loop is completed, login failed, so return false
        return False



# coding=utf8
# Copyright 2020 Cognicept Systems
# Author: Jakub Tomasek (jakub@cognicept.systems)
# --> Configuration class handles Cognicept configuration

from dotenv import dotenv_values
from pathlib import Path
from Crypto.PublicKey import RSA
import subprocess
import os
import re
import requests
import getpass


class Configuration:
    """
    A class to manage agent and Cognicept's configuration
    ...

    Parameters
    ----------
    None

    Methods
    -------    
    load_config(path):
        Loads and parses `runtime.env` file into `object.config` located at `path`.
    configure(args):
        Manages cognicept configuration based on values in `args`.
    save_config():
        Saves configuration into `runtime.env` file.
    get_cognicept_credentials():
        Checks and returns `COGNICEPT_ACCESS_KEY` from configuration.
    get_cognicept_api_uri():
        Checks and returns `COGNICEPT_API_URI` from configuration.
    get_cognicept_user_api_uri():
        Checks and returns `COGNICEPT_USER_APT_URI` from configuration.
    get_field():
        Returns single config value
    configure_ssh():
        Function to setup ssh access for the host machine given user input.
    is_ssh_enabled(args):
        Checks and returns value of `COG_ENABLE_SSH_KEY_AUTH` as bool.
    cognicept_key_rotate:
        Rotates AWS temporary keys retriving them from Cognicept API.
    init_config(args):
        Initiates runtime.env file with values from Cognicept API
    """
    config_path = os.path.expanduser("~/.cognicept/")
    env_path = config_path + "runtime.env"
    _regex_key = r"^([_A-Z0-9]+)$"
    _config_loaded = False

    def load_config(object, path):
        """
        Loads and parses `runtime.env` file into `object.config` located at `path`.

                Parameters:
                        path (str): Cognicept path, e.g. `~/.cognicept/`.                
        """
        object.config_path = os.path.expanduser(path)
        object.env_path = object.config_path + "runtime.env"
        file = Path(object.env_path)

        if ((not file.exists()) or (file.is_dir())):
            print("Configuration file `" + object.env_path + "` does not exist.")
            try:
                with open(object.env_path, 'w') as f:
                    pass
            except:
                print("Failed to initialize the robot: insufficient priviledges to create `/home/username/.cognicept/runtime.env file.")
                return False

        #object.config = dotenv_values(dotenv_path=file.name) if sys.version_info.minor > 5 else dotenv_values(dotenv_path=find_dotenv(), verbose=True)
        object.config = dotenv_values(
            dotenv_path=object.env_path, verbose=True)
        if(len(object.config) == 0):
            print("Configuration file `" + object.env_path +
                  "` is empty or could not be parsed.")
            return False
        object._config_loaded = True
        return True

    def configure(object, args):
        """
        Manages cognicept configuration based on values in `args`:
            * If `args.read` is True, then it prints all configuration,
            * If `args.add` is True, then it will add or modify a single value inputed by user,
            * Otherwise it will iterate through all values and asks for modifications.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`            
        """
        if(args.ssh):
            if object.configure_ssh():
                print(
                    'SSH config done. To apply changes restart agents using `cognicept restart`.')
        elif(args.autocomplete):
            if object.configure_autocomplete():
                print('Autocomplete setup done')
        else:
            if(not object._config_loaded):
                return

            if (not os.access(object.env_path, os.W_OK)):
                print("Error: You don't have writing permissions for `" +
                    object.env_path + "`. Run as `sudo` or change file permissions.")
                return
            if(args.read):
                for key, value in object.config.items():
                    print(key + ': "' + value + '"')
            elif(args.add):
                new_key = ""
                while(new_key == ""):
                    new_key = input("Config name: ")

                    # if empty, exit
                    if(new_key == ""):
                        return
                    # check if matches key specs
                    matches = re.match(object._regex_key, new_key)
                    if matches is None:
                        print(
                            "Error: Key can be uppercase letters, digits, and the '_'. Try again.")
                        new_key = ""

                new_value = ""
                while(new_value == ""):
                    new_value = input("Value: ")
                    if(new_value == ""):
                        return
                    matches = re.match(r"^.*[\"].*$", new_value)
                    if matches is not None:
                        print("Error: Value cannot contain '\"'. Try again.")
                        new_value = ""

                object.config[new_key] = new_value        
            else:
                for key, value in object.config.items():
                    new_value = input(key + "[" + value + "]:")
                    matches = re.match(r"^.*[\"].*$", new_value)
                    if((new_value != "") and (matches == None)):
                        object.config[key] = new_value
            object.save_config()

    def save_config(object):
        """
        Saves configuration into `runtime.env` file

                Parameters:
                    None
        """
        try:
            with open(object.env_path, 'w') as file:
                for key, value in object.config.items():
                    file.write(key + '=' + value + '\n')
        except IOError:
            print("Could not write into `" + object.env_path +
                  "`. Please check write permission or run with `sudo`.")

    def get_cognicept_credentials(object):
        """
        Checks and returns `COGNICEPT_ACCESS_KEY` from configuration.

                Returns:
                    cognicept_access_key (str): value of `COGNICEPT_ACCESS_KEY`
        """
        if "COGNICEPT_ACCESS_KEY" in object.config:
            return object.config["COGNICEPT_ACCESS_KEY"]
        else:
            print('COGNICEPT_ACCESS_KEY missing')

    def get_cognicept_api_uri(object):
        """
        Checks and returns `COGNICEPT_API_URI` from configuration.

                Returns:
                    cognicept_api_uri (str): value of `COGNICEPT_API_URI`, defaults to "https://app.cognicept.systems/api/v1/"
        """
        if "COGNICEPT_API_URI" in object.config:
            return object.config["COGNICEPT_API_URI"]
        else:
            return "https://app.cognicept.systems/api/v1/"

    def get_cognicept_user_api_uri(object):
        """
        Checks and returns `COGNICEPT_USER_API_URI` from configuration.

                Returns:
                    cognicept_user_api_uri (str): value of `COGNICEPT_USER_API_URI`, defaults to "https://app.cognicept.systems/api/v1/"
        """
        if "COGNICEPT_USER_API_URI" in object.config:
            return object.config["COGNICEPT_USER_API_URI"] 
        else:
            return "https://app.cognicept.systems/api/v1/"

    def get_field(object, field_name):
        """
        Returns single config value

                Parameters:
                        field_name (str): name of config field
                Returns:
                        value (str): value
        """
        if field_name in object.config:
            return object.config[field_name]
        else:
            raise KeyError(field_name)

    def _interpret_bool_input(object, input_string):
        """
        Parses user input string bool value 

                Parameters:
                        input_string (str): input value
                Returns:
                        bool_value (bool): True if value is 'Y'
        """
        if(input_string == 'Y'):
            return True
        elif(input_string == 'n'):
            return False
        else:
            return None

    def configure_ssh(object):
        """
        Function to setup ssh access for the host machine given user input

                Parameters:
                        None
                Returns:
                        result (bool): True if some step didn't fail
        """
        print('SSH is used to access the host machine from the isolated docker environment of Cognicept agent.')

        enable_ssh = None
        while(enable_ssh == None):
            enable_ssh = object._interpret_bool_input(
                input("Enable SSH access? (Y/n):"))

        if(not enable_ssh):
            object.config["COG_ENABLE_SSH"] = "False"
            object.config["COG_ENABLE_SSH_KEY_AUTH"] = "False"
            object.config["COG_ENABLE_AUTOMATIC_SSH"] = "False"
            object.save_config()
            return True
        else:
            object.config["COG_ENABLE_SSH"] = "True"

        ssh_authorized_keys_path = os.path.expanduser(
            "~") + "/.ssh/authorized_keys"
        print("\n \nSSH key needs to be used for hosts with disabled password login. "
              "It can also simplify access so you don't need to input the password each time.\n"
              "This process will generate ssh key locally and mount it to the docker container. "
              "The public key is copied to `" + ssh_authorized_keys_path +
              "` to give access.  Root access is needed and you will be prompted for password."
              "The ssh key is neither sent nor stored to the Cognicept server. "
              "If you choose not to, manual password ssh access can be still used.")

        enable_ssh_key = None
        while(enable_ssh_key == None):
            enable_ssh_key = object._interpret_bool_input(
                input("Generate SSH key and give access? (Y/n):"))

        if(not enable_ssh_key):
            object.config["COG_ENABLE_SSH_KEY_AUTH"] = "False"
        else:
            object.config["COG_ENABLE_SSH_KEY_AUTH"] = "True"

        # generate the ssh key and write them in the file
        if object.config["COG_ENABLE_SSH_KEY_AUTH"] == "True":
            cognicet_ssh_directory = object.config_path + "ssh/"
            try:
                if not os.path.exists(cognicet_ssh_directory):
                    os.makedirs(cognicet_ssh_directory)
            except:
                print(
                    "Failed. Don't have privileges to create files/directories within the ssh directory.")
                return False

            # generate the keys
            ssh_key = RSA.generate(2048)
            private_key_path = cognicet_ssh_directory + "id_rsa"
            public_key_path = cognicet_ssh_directory + "id_rsa.pub"
            config_file_path = cognicet_ssh_directory + "config"
            with open(private_key_path, 'wb') as content_file:
                os.chmod(private_key_path, 0o600)
                content_file.write(ssh_key.exportKey('PEM'))
            pub_key = ssh_key.publickey().exportKey('OpenSSH')
            with open(public_key_path, 'wb') as content_file:
                content_file.write(pub_key)
            # add new line at the end of the file
            with open(public_key_path, 'a') as content_file:
                content_file.write("\n")

        default_user = getpass.getuser()
        user_exists = False
        ssh_directory_path = ""

        # retrieve the user name
        user_prompt_number = 0
        while not user_exists:
            user_name = input(
                "Name of the user to access ssh(if empty, defaults to `" + default_user + "`): ")
            if(user_name == ""):
                user_name = default_user
            object.config["COG_SSH_DEFAULT_USER"] = user_name
            
            if(user_name == "root"):
                ssh_directory_path = "/root/.ssh/"
            else:                
                ssh_directory_path = "/home/" + user_name + "/.ssh/"
            if os.path.exists(ssh_directory_path):
                user_exists = True
            else:
                print("User " + user_name +
                      " doesn't seem to exist or openssh server is not installed.")
                user_prompt_number = user_prompt_number + 1
                if(user_prompt_number > 2):
                    print("Failed to find user's openssh configuration directory.")
                    return False

        # copy the keys to authorized_keys file if automatic authentication was enabled
        if object.config["COG_ENABLE_SSH_KEY_AUTH"] == "True":
            authorized_keys_path = ssh_directory_path + "authorized_keys"
            try:
                print('Root access is needed to modify `' +
                      authorized_keys_path + '` and you will be prompted for password.')
                proc = subprocess.call(
                    ['sudo', 'sh', '-c', 'cat ' + public_key_path + ' >> ' + authorized_keys_path])
            except:
                print("Failed! Don't have access to " + authorized_keys_path)
                return

        enable_automatic_ssh = None
        while(enable_automatic_ssh == None):
            enable_automatic_ssh = object._interpret_bool_input(
                input("Enable automatic ssh access? (Y/n):"))

        if(not enable_automatic_ssh):
            object.config["COG_ENABLE_AUTOMATIC_SSH"] = "False"
        else:
            object.config["COG_ENABLE_AUTOMATIC_SSH"] = "True"

        object.save_config()
        return True

    def is_ssh_enabled(object):
        """
        Checks and returns value of `COG_ENABLE_SSH_KEY_AUTH` as bool

                Parameters:
                        None
                Returns:
                        ssh_enabled (bool): True if ssh is enables
        """
        if(not "COG_ENABLE_SSH_KEY_AUTH" in object.config):
            return False

        return object.config["COG_ENABLE_SSH_KEY_AUTH"]


    def fetch_aws_keys(object):
        """
        Fetch AWS temporary keys retriving them from Cognicept API.

                Parameters:
                        
                Returns:
                        result (json): AWS temporary credentials returned from Cognicept API
        """
        try:
            headers = {
                'Authorization': 'Basic ' + object.get_cognicept_credentials()
            }
            resp = requests.get(object.get_cognicept_api_uri(
            ) + 'aws/assume_role', headers=headers, timeout=5)
            if resp.status_code != 200:
                print('Login error: wrong credentials.')
                return False
            return resp.json()            
        except requests.exceptions.Timeout:
            print("Cognicept REST API error: time out.")
            return False
        except requests.exceptions.TooManyRedirects:
            print("Cognicept REST API error: Wrong endpoint.")
            return False
        except Exception as e:
            print("Cognicept REST API error" + str(e))
            return False
        object.save_config()

        return True

    def cognicept_key_rotate(object, args):
        """
        Rotates AWS temporary keys retriving them from Cognicept API.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """
        print("Updating cloud credentials.")
        try:
            response = object.fetch_aws_keys()
            if response==False:
                print("Error connecting Cognicept API. Please contact support@cognicept.systems")
                return False
            object.config["AWS_ACCESS_KEY_ID"] = response[
                "AccessKeyId"]
            object.config["AWS_SECRET_ACCESS_KEY"] = response[
                "SecretAccessKey"]
            object.config["AWS_SESSION_TOKEN"] = response[
                "SessionToken"]
            object.config["AWS_TOKEN_EXPIRATION"] = response[
                "Expiration"]
            print("Cloud access keys rotated successfully!")
        except requests.exceptions.Timeout:
            print("Cognicept REST API error: time out.")
            return False
        except requests.exceptions.TooManyRedirects:
            print("Cognicept REST API error: Wrong endpoint.")
            return False
        except Exception as e:
            print("Cognicept REST API error" + str(e))
            raise SystemExit()
        object.save_config()

        return True

    
    def init_config(object, args):
        """
        Initiates runtime.env file with values from Cognicept API.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        result (bool): True if succeeded
        """
        username = input('Username: ')
        password = getpass.getpass()
        api_uri = object.get_cognicept_user_api_uri()
        try:
            x = requests.post(api_uri + 'user/login', json={'username': username, 'password': password})
            if 'access_token' not in x.json():
                print('Failed to initialize the robot: Wrong credentials') 
            else:
                auth_key = x.json()["access_token"]

                r = requests.get(api_uri + 'spinup/config/' + args.org_id + '/' + args.robot_id, headers={'Authorization': 'Bearer {}'.format(auth_key)})
                j = r.json()
                if j is not None:
                    for key in j:
                        object.config[key] = j[key]
                    object.save_config()
                    print("Successfully initialized configuration for the robot `" + args.robot_id + "`. To start agents run `cognicept start`")
                    return True
                else:
                    print("Failed to initialize the robot: ID `" + args.robot_id + "` in organization `" + args.org_id + "` not found")
                    return False
        except requests.exceptions.Timeout:
            print("Cognicept REST API error: time out.")
            return False
        except requests.exceptions.TooManyRedirects:
            print("Cognicept REST API error: Wrong endpoint.")
            return False
        except Exception as e:
            print("Cognicept REST API error" + str(e))
            raise SystemExit()

    def configure_autocomplete(object):
        """

        Function to setup autcomplate functionality
                Parameters:
                        None
                Returns:
                        result (bool): True if some step didn't fail

        """

        enable_autocomplete = None
        while(enable_autocomplete == None):
            enable_autocomplete = object._interpret_bool_input(
                input("Enable Autocomplete? (Y/n):"))
        
        if (not enable_autocomplete):
            print("Autocomplete was not configured")
            return True

        bash_complete_path = "/etc/bash_completion.d/"
        tmp_path = os.path.expanduser("~") + "/.cognicept/"
        
        print("Root access is needed to place bash completition script at `" + bash_complete_path +
              "' you will be prompted for password.")

        allow_root_access = None
        while(allow_root_access == None):
            allow_root_access = object._interpret_bool_input(
            input("Generate script and give access? (Y/n):"))
        subprocess.call(['bash', '-c', 'activate-global-python-argcomplete --dest ' + tmp_path])
        subprocess.call(['sudo', 'bash', '-c', 'mv ' + tmp_path + "python-argcomplete" + " /etc/bash_completion.d/"])
        print("Please restart the terminal.....")
        return True



# coding=utf8
# Copyright 2020 Cognicept Systems
# Author: Swarooph Seshadri (swarooph@cognicept.systems)
# --> RosbagRecord class handles on demand rosbag recording for cognicept-shell

from dotenv import dotenv_values
import os
import time
from datetime import datetime, timedelta
import glob
import docker
from cogniceptshell.common import bcolors
from cogniceptshell.agent_life_cycle import AgentLifeCycle
from cogniceptshell.pusher import Pusher
from cogniceptshell.common import permission_safe_docker_call

class RosbagRecord:
    """
    A class to manage Cognicept's rosbag recording sessions
    ...

    Parameters
    ----------
    None

    Methods
    -------
    record(args):
        Main entrypoint to manage recording session for start/all/stop/pause/resume/status modes.
    start_record(args):
        Executes a docker command inside `cgs_bagger_server` container to start the recording session.
        Has built in retry mechanism based on bag file status which when fails restarts the `cgs_bagger_server` container automatically. 
        `args.start` should have a mandatory list of topics to record or `args.all` to record all topics.
    stop_record(args):
        Executes a docker command inside `cgs_bagger_server` container to stop the recording session.
        Has built in retry mechanism based on bag file status which when fails restarts the `cgs_bagger_server` container automatically.
        If `args.stop` is set to `autopush`, it calls a `Pusher` instance to upload the latest bag to S3.
    pause_record(args):
        Executes a docker command inside `cgs_bagger_server` container to pause the recording session.
        Has built in retry mechanism based on the container response in case command fails.
    resume_record(args):
        Executes a docker command inside `cgs_bagger_server` container to resume the recording session.
        Has built in retry mechanism  based on the container response in case command fails.
    get_record_status(args):
        Executes a docker command inside `cgs_bagger_server` container to get the status the recording session.
        Has built in retry mechanism  based on the container response in case command fails.
    print_status(args):
        Parses container response and prints to console if one exists and returns TRUE. If not returns false.
        Used to manage retry mechanism for pause/resume/status modes.
    get_latest_type_bag(args):
        Utility method to get the latest '.bag.active' or '.bag' file depending on what `type` is set to.
        Used to manage retry mechanism for start/stop modes.
    restart_bag_record_server(args):
        Utility function to restart the `cgs_bagger_server` container.
        Uses an instance of the `AgentLifeCycle` class.
    """

    def record(self, args):
        """
        Main entrypoint to manage recording session for start/all/stop/pause/resume/status modes.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
        """
        # Entry point method to decide which mode to call
        if((args.start is None) and (args.stop is False)
           and (args.pause is False) and (args.resume is False)
           and (args.all is False) and (args.status is False)):
            # start or stop must be provided
            print(bcolors.FAIL +
                  ("Required command is missing."
                   " Check `cognicept record --help` "
                   "for more commands available.") +
                  bcolors.ENDC)
        elif args.start:
            # to start recording
            print(bcolors.OKBLUE + "Starting recording" + bcolors.ENDC)
            self.start_record(args)
        elif args.all:
            # to start recording
            print(bcolors.OKBLUE + "Starting recording" + bcolors.ENDC)
            self.start_record(args)
        elif args.stop:
            # to stop recording
            print(bcolors.OKBLUE + "Stopping recording" + bcolors.ENDC)
            self.stop_record(args)
        elif args.pause:
            # to pause recording
            print(bcolors.OKBLUE + "Pausing recording" + bcolors.ENDC)
            self.pause_record(args)
        elif args.resume:
            # to resume recording
            print(bcolors.OKBLUE + "Resuming recording" + bcolors.ENDC)
            self.resume_record(args)
        elif args.status:
            # query recording status
            print(bcolors.OKBLUE + "Getting recording status" + bcolors.ENDC)
            self.get_record_status(args)
        else:
            # This should never execute
            print(bcolors.FAIL +
                  ("Required command is missing."
                   " Check `cognicept record --help` "
                   "for more commands available.") +
                  bcolors.ENDC)

    def validate_topic_names(self, args):
        valid_topic_names = []
        invalid_topic_names = []
        for topic in args.start:
            if topic.startswith('/'):
                valid_topic_names.append(topic)
            else:
                invalid_topic_names.append(topic)
                        
        if not valid_topic_names:
            print(bcolors.FAIL +
                  ("Valid Topic Names should start with a `/`. "
                  "Only these will be considered for recording. "
                  "No valid topic names found.") +
                  bcolors.ENDC)
            raise SystemExit
        else:
            print('Valid Topic Names: ', ','.join(valid_topic_names))

        args.start = valid_topic_names

    def start_record(self, args):
        """
        Executes a docker command inside `cgs_bagger_server` container to start the recording session.
        Has built in retry mechanism based on bag file status which when fails restarts the `cgs_bagger_server` container automatically. 
        `args.start` should have a mandatory list of topics to record 
        (or)
        `args.all` to record all topics.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
        """
        rosbag_start_cmd = "record_cmd.bash start "
        if args.all:
            selected_topics = '-a'
            rosbag_start_cmd = rosbag_start_cmd + selected_topics
        else:
            self.validate_topic_names(args)
            selected_topics = ','.join(args.start)
            rosbag_start_cmd = rosbag_start_cmd + "'" + selected_topics + "'"

        permission_docker_call_result = permission_safe_docker_call(docker.from_env)
        if permission_docker_call_result is None:
            return False
        client = permission_docker_call_result
        try:
            container = client.containers.get('cgs_bagger_server')
            num_retries = 3

            for retry in range(0, num_retries):

                container_response = container.exec_run(
                    rosbag_start_cmd, stream=True, stdout=True)
                # status = self.print_status(container_response)
                latest_active_bag = self.get_latest_type_bag(args, "active")

                if latest_active_bag:
                    print("Active bag file:   ", latest_active_bag)
                    break

                if retry < num_retries - 1:
                    print(bcolors.WARNING +
                          "Try failed. Retrying..." + bcolors.ENDC)
                else:
                    print(
                        bcolors.FAIL +
                        ("Recording Failed. "
                         "Restarting bag recording server. ") +
                        bcolors.ENDC)
                    self.restart_bag_record_server(args)
                    print(
                        bcolors.OKBLUE +
                        "Please try your record command again." +
                        bcolors.ENDC)
                    raise(SystemExit)

            print(bcolors.OKBLUE +
                  ("Use `cognicept record --status` "
                   "for detailed progress.\n"
                   "Use `cognicept record --stop` "
                   "to stop recording.") +
                  bcolors.ENDC)
        except docker.errors.APIError as e:
            print(
                bcolors.FAIL + "Bag Record Server not running or incorrect command provided." + bcolors.ENDC)

    def stop_record(self, args):
        """
        Executes a docker command inside `cgs_bagger_server` container to stop the recording session.
        Has built in retry mechanism based on bag file status which when fails restarts the `cgs_bagger_server` container automatically.
        If `args.stop` is set to `autopush`, it calls a `Pusher` instance to upload the latest bag to S3.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
        """
        rosbag_stop_cmd = "record_cmd.bash stop"
        permission_docker_call_result = permission_safe_docker_call(docker.from_env)
        if permission_docker_call_result is None:
            return False
        client = permission_docker_call_result
        try:
            container = client.containers.get('cgs_bagger_server')
            num_retries = 3

            for retry in range(0, num_retries):

                container_response = container.exec_run(
                    rosbag_stop_cmd, stream=True, stdout=True)
                # status = self.print_status(container_response)
                latest_bag = self.get_latest_type_bag(args, "bag")

                if latest_bag:
                    print("Recorded bag file: ", latest_bag)
                    break

                if retry < num_retries - 1:
                    print(bcolors.WARNING +
                          "Try failed. Retrying..." + bcolors.ENDC)
                else:
                    print(
                        bcolors.FAIL +
                        ("Unable to stop recording. "
                         "Restarting bag recording server. ") +
                        bcolors.ENDC)
                    self.restart_bag_record_server(args)
                    print(
                        bcolors.OKBLUE +
                        "Please start a new recording session." +
                        bcolors.ENDC)
                    raise(SystemExit)

            if args.stop == "autopush":
                pusher_instance = Pusher()
                args.bag = "latest"
                pusher_instance.push(args)
            else:
                print(bcolors.OKBLUE +
                      ("Use `cognicept push --bag` "
                       "to upload the latest recording to cloud.") +
                      bcolors.ENDC)
        except docker.errors.APIError as e:
            print(
                bcolors.FAIL + "Bag Record Server not running or incorrect command provided." + bcolors.ENDC)

    def pause_record(self, args):
        """
        Executes a docker command inside `cgs_bagger_server` container to pause the recording session.
        Has built in retry mechanism based on the container response in case command fails.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
        """
        rosbag_pause_cmd = "record_cmd.bash pause"
        permission_docker_call_result = permission_safe_docker_call(docker.from_env)
        if permission_docker_call_result is None:
            return False
        client = permission_docker_call_result
        try:
            container = client.containers.get('cgs_bagger_server')
            num_retries = 3

            for retry in range(0, num_retries):

                container_response = container.exec_run(
                    rosbag_pause_cmd, stream=True, stdout=True)
                resp_status = self.print_status(container_response)

                if resp_status:
                    print(bcolors.OKBLUE +
                          ("Use `cognicept record --resume` "
                           "to resume recording session.") +
                          bcolors.ENDC)
                    break

                if retry < num_retries - 1:
                    print(bcolors.WARNING +
                          "Try failed. Retrying..." + bcolors.ENDC)
                else:
                    print(
                        bcolors.FAIL +
                        ("Pausing Failed. "
                         "Try to stop completely using `cognicept record --stop`") +
                        bcolors.ENDC)

        except docker.errors.APIError as e:
            print(
                bcolors.FAIL + "Bag Record Server not running or incorrect command provided." + bcolors.ENDC)

    def resume_record(self, args):
        """
        Executes a docker command inside `cgs_bagger_server` container to resume the recording session.
        Has built in retry mechanism  based on the container response in case command fails.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
        """
        rosbag_resume_cmd = "record_cmd.bash resume"
        permission_docker_call_result = permission_safe_docker_call(docker.from_env)
        if permission_docker_call_result is None:
            return False
        client = permission_docker_call_result
        try:
            container = client.containers.get('cgs_bagger_server')
            num_retries = 3

            for retry in range(0, num_retries):

                container_response = container.exec_run(
                    rosbag_resume_cmd, stream=True, stdout=True)
                resp_status = self.print_status(container_response)

                if resp_status:
                    print(bcolors.OKBLUE +
                          ("Use `cognicept record --stop` "
                           "to stop recording session.") +
                          bcolors.ENDC)
                    break

                if retry < num_retries - 1:
                    print(bcolors.WARNING +
                          "Try failed. Retrying..." + bcolors.ENDC)
                else:
                    print(
                        bcolors.FAIL +
                        ("Resuming Failed. "
                         "Try to stop completely using `cognicept record --stop`") +
                        bcolors.ENDC)

        except docker.errors.APIError as e:
            print(
                bcolors.FAIL + "Bag Record Server not running or incorrect command provided." + bcolors.ENDC)

    def get_record_status(self, args):
        """
        Executes a docker command inside `cgs_bagger_server` container to get the status the recording session.
        Has built in retry mechanism  based on the container response in case command fails.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
        """
        rosbag_status_cmd = "record_cmd.bash status"
        permission_docker_call_result = permission_safe_docker_call(docker.from_env)
        if permission_docker_call_result is None:
            return False
        client = permission_docker_call_result
        try:
            container = client.containers.get('cgs_bagger_server')
            num_retries = 3

            for retry in range(0, num_retries):

                container_response = container.exec_run(
                    rosbag_status_cmd, stream=True, stdout=True)
                resp_status = self.print_status(container_response)

                if resp_status:
                    print(bcolors.OKBLUE +
                          ("Use `cognicept record --help` "
                           "for more information.") +
                          bcolors.ENDC)
                    break

                if retry < num_retries - 1:
                    print(bcolors.WARNING +
                          "Try failed. Retrying..." + bcolors.ENDC)
                else:
                    print(
                        bcolors.FAIL +
                        ("Failed to get status. "
                         "Try to stop completely using `cognicept record --stop` and restart bag recording server.") +
                        bcolors.ENDC)

        except docker.errors.APIError as e:
            print(
                bcolors.FAIL + "Bag Record Server not running or incorrect command provided." + bcolors.ENDC)

    def print_status(self, container_response):
        """
        Parses container response and prints to console if one exists
        Returns True/False based on if successful response was achieved.
        Used to manage retry mechanism for pause/resume/status modes.

                Parameters:
                        container_response: populated argument namespace returned by `argparse.parse_args()`
                Returns:
                        response_status (boolean): True if container_response is valid
        """
        output_response_str = str(next(container_response.output))
        str_beg = output_response_str.rfind('Recording state:')
        str_end = output_response_str.rfind('Goal Succeeded')

        if str_beg == -1:
            return False
        else:
            print(output_response_str[str_beg:str_end-2])
            return True

    def get_latest_type_bag(self, args, search_type):
        """
        Get the latest '.bag.active' or '.bag' file depending on what `type` is set to.
        Used to manage retry mechanism for start/stop modes.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
                        search_type: Either `.active` or `.bag` based on which type of file needs to be found
                Returns:
                        latest_file (str): If file is found, full path str of the file otherwise ""
        """
        # Wait for the OS to initialize an active bag
        time.sleep(7.0)
        bag_path = os.path.expanduser(args.path+"bags")
        bag_list = glob.glob(bag_path+"/*." + search_type)
        if bag_list:
            latest_file = max(bag_list, key=os.path.getctime)
            latest_timestamp = os.path.getctime(latest_file)
            file_time = datetime.fromtimestamp(
                latest_timestamp)  # .strftime('%Y-%m-%d %H:%M:%S')
            sys_time = datetime.now()  # .strftime('%Y-%m-%d %H:%M:%S')
            if sys_time - file_time > timedelta(10.0):
                print('File time: ', file_time)
                print('Sys time: ', sys_time)
                print('Latest found file is not recent enough.')
                return ""
            else:
                # print('File time: ', file_time)
                # print('Sys time: ', sys_time)
                return latest_file
        else:
            return ""

    def restart_bag_record_server(self, args):
        """
        Utility function to restart the `cgs_bagger_server` container.
        Uses an instance of the `AgentLifeCycle` class.

                Parameters:
                        args: populated argument namespace returned by `argparse.parse_args()`
        """
        agent_lifetime = AgentLifeCycle()
        args.agents = None
        args.list = ['cgs_bagger_server']
        agent_lifetime.restart(args)
